%% Shlukovn
% V tomto cvien se ble podvme na jeden ze zkladnch algoritm pro
% shlukovn, na algoritmus k-means. Naimplementujeme jej a aplikujeme na
% loze segmentace barevnch bod v obrzku.

%% Opakovn
% Zopakujte si, co vte o algoritmu k-means z pednky. Jak funguje?
%
% Jednm z kol z minulch cvien bylo implementovat klasifiktor pro
% metodu nejbliho souseda. Tuto funkci s vhodou vyuijeme v algoritmu
% k-means. Funkce by mohla vypadat teba takto:
dbtype pred1NearestNeighbor

%%
% Sprvnou funkci naeho algoritmu si ukeme na jednoduchm problmu.
% Vygenerujeme 2D data z intervalu (0,1)^2, urme dva centroidy, teba
% jako (0,0) a (1,0.5) a znzornme, ke ktermu centroidu algoritmus body
% piadil.
data = rand(2,500);
centroids = [0 0; 1 0.5]';
cl = pred1NearestNeighbor(centroids, data);
plotGroupedData(data, cl); axis equal;

%% Algoritmus k-means
% Zopakujme si, jak algoritmus k-means pracuje. Pravideln std dv fze:
% piazen bod k centroidm a pravu pozic centroid.
%
% # Inicializuj polohu centroid (nhodn, nebo vyber _k_ nhodnch bod z trnovac sady)
% # Pia trnovac body k centroidm
% # Uprav pozici centroid na zklad piazench bod
% # Pokud se pozice centroid (nebo piazen bod do shluk) od posledn
% iterace nezmnila, ukoni algoritmus
%
% Implementace algoritmu me vypadat nap. takto:
dbtype clustKMeans

%%
% Algoritmus si opt meme ovit na umlch datech. Vytvome z dat dva
% tvereky - jeden v intervalu (0,1)^2, druh v intervalu (1,2)^2.
% Pokusme se najt 2 shluky.
data2 = [data 1+data];
[centroids, cl] = clustKMeans(2,data2);
plotGroupedData(data2, cl); hold on;
plot(centroids(1,:), centroids(2,:), 'ko', ...
    'MarkerSize', 10, 'Color', 'k', 'MarkerFaceColor', 'k');
axis equal;

%%
% Pokusme-li se vytvoit vce shluk, situace bude vypadat nsledovn.
[centroids, cl] = clustKMeans(3,data2);
plotGroupedData(data2, cl); hold on;
plot(centroids(1,:), centroids(2,:), 'ko', ...
    'MarkerSize', 10, 'Color', 'k', 'MarkerFaceColor', 'k');
axis equal;

[centroids, cl] = clustKMeans(4,data2);
plotGroupedData(data2, cl); hold on;
plot(centroids(1,:), centroids(2,:), 'ko', ...
    'MarkerSize', 10, 'Color', 'k', 'MarkerFaceColor', 'k');
axis equal;

%%
% Pro 4 shluky se velice asto projev nalezen loklnho optima, viz
% obrzek. Dokete vysvtlit pro? Jak by podle vs mla vypadat situace,
% kdyby bylo nalezeno globln optimum? Jak by se dal n algoritmus
% upravit, aby toto optimum nachzel spolehlivji?

%% Segmentace obrzku
% Pokusme se rozdlit body v obrzku do skupin, a to nkolika zpsoby.
% Budeme brt v vahu
%
% # pouze barevn sloky bodu
% # pouze souadnice bod
% # barevn sloky i souadnice bod, s rznmi vhami
%
% Podvejme se na pouit obrzek:
pic = imread('lekniny.jpg','jpg');
[Ny,Nx,foo] = size(pic);
figure; imshow(pic);
%%
% Pevedeme jej do formy, kter umouje zpracovat body
% algoritmem k-means. Znamen to vytvoit matici velikosti [5 x (nx*ny)],
% kde pro kad bod obrzku, kterch je (nx*ny), vytvome 5 souadnic: R,
% G a B barevn sloky a x- a y-ovou pozici bodu v obrzku.
[pts, ny, nx] = transformPicture(pic);

%%
% Ped segmentac (shlukovnm) se obvykle provd normalizace pznak tak,
% aby mly vechny stejn rozsah. V naem ppad se barevn sloky
% pohybuj v hodnotch 0-255 a souadnice jsou v rozsahu 0-400, resp.
% 0-300. Jsou tedy soumiteln a normalizaci nen teba provdt.
%
%% Segmentace barev
% Shlukovou analzu meme vyut pro redukci potu barev v obrzku (ztrtov komprese). V
% klasickm nekomprimovanm formtu jsou pro kad bod obrzku uvedeny
% vechny ti sloky barvy na 8 bitech, tj. 3 byty. Velikost souboru je pak
% piblin rovna 
%
% $$\mathrm{velikost} = 3 \cdot N_x \cdot N_y$$
%
% bajt, kde _Nx_ _x_-ov rozmr obrzku a _Ny_ je _y_-ov rozmr obrzku.
% N nekomprimovan obrzek m tedy velikost v bajtech

velikostnekompr = 3*Nx*Ny

%%
% Pokud ale zredukujeme poet barev _Nc_ nap. jen na 16, sta pak u kadho bodu
% jen urit index barvy, kterou m pout. Grafick soubor pak bude
% obsahovat 3bytovou definici kad z _Nc_ barev a log(_Nc_)-bitov index
% barvy pro kad bod, take
%
% $$\mathrm{velikost} = 3 \cdot N_c + N_x \cdot N_y \cdot \frac{log_2(N_c)}{8}$$
%
% Pouijeme-li tedy 16 barev, bude mt n obrzek velikost
Nc = 16;
velikost16barev = 3*Nc + Nx*Ny*log2(Nc)/8
%%
% Zkusme tedy, jak bude vypadat n obrzek pot, co vytvome shluky bod
% v prostoru barev (souadnice budeme ignorovat) a barvy kadho bodu
% nahradme barvou pslunho centroidu.
ptsC = pts(1:3,:);
% [centroids, cl] = clustKMeans(Nc, ptsC) % n algoritmus
% [model, cl] = cmeans(ptsC, Nc);  centroids = model.X; % STPRtool
[cl, centroids] = kmeans(ptsC', Nc); cl = cl'; centroids = centroids'; % Statistics toolbox
pts1 = replaceColorsWithCentroids(pts, cl);

%%
% Vytvome obrzek a zobrazme ho.
pic1 = transformPictureBack(pts1, ny, nx);
figure; imshow(pic1);

%% Segmentace souadnic
% V minul analze jsme ignorovali souadnice bodu v obrzku. Jak se
% analza zmn, kdy budeme pro zmnu brt v vahu jen souadnice bod?

ptsC = pts(4:5,:);
% [centroids, cl] = clustKMeans(Nc, ptsC)
[cl, centroids] = kmeans(ptsC', Nc); cl = cl'; centroids = centroids';
pts2 = replaceColorsWithCentroids(pts, cl);
pic2 = transformPictureBack(pts2, ny, nx);
figure; imshow(pic2);

%%
% Rozdl v obou ppadech je samozejm obrovsk. Kdy jsme shlukovali
% barvy, byl obrzek pomrn vrn, protoe shluky mohly bt prostorov
% nesouvisl. Kdy jsme shlukovali podle souadnic, shluky jsou kompaktn,
% ale obrzek si moc podobn nen.

%% Segmentace objekt v obrzku
% Zkombinujeme-li oba pstupy, mli bychom dostat metodu, kter do jednoho
% shluku zaad barevn podobn body, kter zrove le blzko sebe. 

Nc = 16;
ptsC = pts;
ptsC(4:5,:) = 0.8*ptsC(4:5,:);
% [centroids, cl] = clustKMeans(Nc, ptsC)
[cl, centroids] = kmeans(ptsC', Nc); cl = cl'; centroids = centroids';
pts3 = replaceColorsWithCentroids(pts, cl);
pic3 = transformPictureBack(pts3, ny, nx);
figure; imshow(pic3);

%%
% V obrzk leknn se nm touto metodou podailo docela dobe
% identifikovat oblasti, kde jsou zobrazeny kvty, kde plovouc listy a kde
% prosvt hladina nebo kde jsou stny.